﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Collections;
using System.IO;
using System.Diagnostics;
using System.Windows.Forms;

namespace KyotoUnivEngHywrExecutionEnvironment
{
    class Parameter_Identification
    {
        /// <summary>
        /// パラメタ同定の本体
        /// </summary>
        /// <param name="x_value">パラメタ値</param>
        /// <param name="para_info">パラメタ名など</param>
        /// <param name="hymco_dir">hymco.exeのディレクトリ</param>
        /// <param name="model_pass">モデル名、パラメタ数、パラメタ名</param>
        /// <param name="dataGridView1">パラメタ表</param>
        /// <param name="filename">構造定義ファイル名</param>
        /// <param name="columns">グループ番号が設定されている列番号</param>
        /// <param name="rows">グループ番号が設定されている行番号</param>
        public void parameter_identify(ArrayList x_value, ArrayList para_info, string hymco_dir, ArrayList model_pass, DataGridView dataGridView1,
            string filename, string[] columns, string[] rows)
        {
            // グループ内のパラメータが同じか確認する
            for (int i = 0; i < columns.Count(); i++)
            {
                if (columns[i] != null)
                {
                    string[] check_column = columns[i].Split('\t');

                    for (int j = 0; j < check_column.Count() - 1; j++)
                    {
                        if (check_column[0] != check_column[j])
                        {
                            MessageBox.Show("Please unify parameters in the group", "Error");
                            goto Finish;
                        }
                    }
                }
            }            
            
            // 初期設定
            Form2 Set_para = new Form2(x_value, para_info);
            Set_para.ShowDialog();

            // 変数を読み込んでいるかのチェック
            if (Set_para.check_ok == false)
            {
                goto Finish;
            }            
            
            // 変数の定義と読み込み
            int cal_count = Set_para.Cal_count;
            ArrayList x_min = new ArrayList();
            ArrayList x_init = new ArrayList();
            ArrayList x_max = new ArrayList();
            x_min = Set_para.Minimum_data();
            x_init = Set_para.Initial_data();
            x_max = Set_para.Maximum_data();
            string filename_cal = Set_para.Filename_of_cal;
            string filename_obs = Set_para.Filename_of_obs;
            string model_name = Set_para.Modelname;
            string ob_func = Set_para.Ob_func;
            string time_type = Set_para.Time_type;
            bool check_out = Set_para.Check_out;

            // SCE-UA 起動に向けての準備
            // 初期値ファイルを作成
            Directory.SetCurrentDirectory(hymco_dir); // 現在の作業ディレクトリを hymco_dir に設定する
            using (StreamWriter write_init = new StreamWriter("cmpeeInitParam.dat"))
            // SCE-UA.exe に与える初期パラメータの値を cmpeeInitParam.dat に記述する
            {
                // パラメータの個数
                write_init.WriteLine(x_min.Count);

                //パラメータの最小値を出力
                for (int i = 0; i < x_min.Count; i++)
                {
                    write_init.WriteLine(x_min[i]);
                }

                //パラメータの初期値を出力
                for (int i = 0; i < x_init.Count; i++)
                {
                    write_init.WriteLine(x_init[i]);
                }

                //パラメータの最大値を出力
                for (int i = 0; i < x_max.Count ; i++)
                {
                    write_init.WriteLine(x_max[i]);
                }
                write_init.Close();
            }

            // 最初に cmpeeUpdateParam.dat と cmpeeUpdateParamTmp.dat を消しておく
            File.Delete("cmpeeUpdateParam.dat");
            File.Delete("cmpeeUpdateParamTmp.dat");

            // cmpeeCount.dat を作る。このファイルには SCE-UA の繰り返し回数を記録する。
            // このファイルは SCE-UA.exe で更新される。
            using (StreamWriter cs2for = new StreamWriter("cmpeeCount.dat"))
            {
                cs2for.WriteLine("0"); // 初期値として 0 を書き込む
                cs2for.Close();
            }

            // SCE-UAの起動
            if (File.Exists("SCE-UA.exe") == false)
            {
                MessageBox.Show("There is no SCE-UA.exe","Error");
                goto Finish;
            }
            Process cmd_pro_pi = new Process();
            cmd_pro_pi.StartInfo.FileName = "cmd.exe"; // コマンドプロンプトの文字列設定
            cmd_pro_pi.StartInfo.Arguments = "/K " + "SCE-UA.exe"; // SCE-UA の実行の文字列設定
            cmd_pro_pi.StartInfo.UseShellExecute = false;
            cmd_pro_pi.StartInfo.CreateNoWindow = true;
            cmd_pro_pi.StartInfo.RedirectStandardOutput = true;
            
            // コマンドラインでの SCE-UA.exe の実行。SCE-UA.exe はここで起動されて
            // 所定の回数を繰り返すまで、終了しない。
            cmd_pro_pi.Start();

            ArrayList buf_true = new ArrayList();
            string line;

            // 計算結果表示開始
            Form3 display_result = new Form3(); // 結果表示用フォームのオブジェクトを生成する
            display_result.Show(); // モードレスフォームとしてフォームを表示する
            string list = "Parameter Identification Start";

            display_result.textDisplay0(list); // 文字列表示

            list = "Count             objective             ";
            for (int i = 0; i < x_init.Count; i++)
            {
                list += "No." + i + " Group " +  "              ";
            }
            display_result.textDisplay0(list); // 文字列表示

            // 1回目の解析準備
            // dataGridView1に初期値を代入する
            for (int i = (int)model_pass[1]; i < dataGridView1.Columns.Count; i++)
            {
                for (int j = 0; j < dataGridView1.Rows.Count-1; j++)
                {
                    if (dataGridView1[i, j].Value.ToString() != "*")
                    {
                        int group_num = int.Parse(((dataGridView1[i, j]).Value).ToString());
                        dataGridView1[(i - (int)(model_pass[1])), j].Value = x_init[group_num].ToString();
                    }
                }
            }

            // ファイルの保存
            FileIO io = new FileIO();
            ArrayList buf_out = new ArrayList();
            ArrayList buf_datagridview = new ArrayList();
            buf_out = io.file_in(filename, 1, model_pass);
            buf_datagridview = io.read_datagrid(dataGridView1,1);

            //ファイル名作成(ディレクトリーはそのまま使う)
            int dummy_filename_count = filename.Split('\\').Count();
            string dummy_filename = filename.Split('\\')[dummy_filename_count-1];
            string filename_out = filename.Replace(dummy_filename, "Parameter_identification.xml");

            // 構造定義ファイルを更新する。
            io.file_out(filename, filename_out, buf_out, buf_datagridview, model_pass);
            
            // 1回目解析開始。以下の関数で hymco.exe が呼び出される。
            read_calculate re_cal = new read_calculate();
            re_cal.calculate("Parameter_identificationProjectFile.xml", hymco_dir);
 
            // 目的関数の結果取得
            objective_function op_func = new objective_function(filename_cal, filename_obs);
            string true_value_func = op_func.ob_cal(ob_func, time_type)[0].ToString();
            double dummy_value_func = double.Parse(op_func.ob_cal(ob_func, time_type)[1].ToString());

            // 目的関数がベストかチェックする(1回目)
            display_result.best(dummy_value_func, 0);

            // SCE-UA.exe に目的関数の値を渡すために、ファイルに初回の目的関数の値を書く。
            using (StreamWriter write_cs2for = new StreamWriter("cmpeeObjVal 0.dat"))
            {
                write_cs2for.WriteLine(dummy_value_func);
                write_cs2for.Close();
            }

            // 結果の表示(1回目)
            list = "1          " + true_value_func + "       ";
            for (int i = 0; i < x_init.Count; i++)
            {
                list += x_init[i].ToString() + "       ";
            }
            list += "1"; // 2回目以降と合わせるためダミー値を入れる("1")
            display_result.textDisplay1(list);


            // 2回目以降の解析
            for (int i = 1; i < cal_count; i++)
            {
                // cmpeeUpdateParam.dat には SCE-UA.exe によって推薦される新しいパラメータ値が追加書きされる。
                //ファイルが存在しない、もしくは書き込み不可能ならばここで足止め
                // 
                while (File.Exists("cmpeeUpdateParam.dat") == false)
                {
                }

            Again: ; // "cmpeeUpdateParam.dat が更新されたかをチェックする。
                File.Copy("cmpeeUpdateParam.dat", "cmpeeUpdateParamTmp.dat");
                // cmpeeUpdateParam.dat は SCE-UA.exe によって更新される。SCE-UA.exe と作業が重ならないように
                // 一旦、cmpeeUpdateParamTmp.dat にコピーして、ファイルが更新されたかをチェックする。

                ArrayList buf = new ArrayList();

                using (StreamReader r = new StreamReader("cmpeeUpdateParamTmp.dat", System.Text.Encoding.GetEncoding("shift_jis")))
                {
                    while ((line = r.ReadLine()) != null)
                    {
                        buf.Add(line); // cmpeeUpdateParamTmp.dat をリストに読み取る
                    }
                    r.Close();
                }
                File.Delete("cmpeeUpdateParamTmp.dat");
                // MessageBox.Show("","check"); //チェック用

                // パラメータ値を更新
                int bc = buf.Count; // cmpeeUpdateParamTmp.dat の行数のカウント
                int btc = buf_true.Count; 
                int count = bc - btc;

                if (bc < i + 1)
                {
                    goto Again; // cmpeeUpdateParamTmp.dat が更新されていないので戻る。
                }
                
                for (int j = count; j > 0; j--)
                {
                    buf_true.Add(buf[bc - j]); // buf の最終行を buf_true に入れる。
                }

                // xmlの更新と計算を行う
                // 新しいパラメータを取得
                string[] new_para = buf_true[i].ToString().Split(' ', '\t');

                // dataGridView1 (機能1のテーブル) に新しいパラメータ値を代入
                for (int j = (int)model_pass[1]; j < dataGridView1.Columns.Count; j++)
                {
                    for (int k = 0; k < dataGridView1.Rows.Count - 1; k++)
                    {
                        if (dataGridView1[j, k].Value.ToString() != "*")
                        {
                            int group_num = int.Parse(((dataGridView1[j, k]).Value).ToString());
                            dataGridView1[(j - (int)(model_pass[1])), k].Value = new_para[group_num+1];
                        }
                    }
                }

                // 構造定義ファイルの保存開始
                buf_out.Clear(); // 以前の構造定義ファイルの情報を消去する
                buf_datagridview.Clear(); 
                buf_out = io.file_in(
                    filename, // 構造定義ファイルの名前 Parameter_identification.xml
                    1, // 出力時は 1 を指定
                    model_pass // パラメータの値の情報が入ったリスト
                    ); // 
                buf_datagridview = io.read_datagrid(dataGridView1, 1); // 更新した dataGridView1 を読み取る

                // buf_out と buf_datagridview を照らし合わせて、構造定義ファイルを更新する。
                io.file_out(filename, filename_out, buf_out, buf_datagridview, model_pass);

                // CommonMP による流出計算
                re_cal.calculate("Parameter_identificationProjectFile.xml", hymco_dir);

                //計算結果を出力
                objective_function ob_funcs = new objective_function(filename_cal, filename_obs);
                true_value_func = ob_funcs.ob_cal(ob_func, time_type)[0].ToString();
                dummy_value_func = double.Parse(ob_funcs.ob_cal(ob_func, time_type)[1].ToString());

                //目的関数がベストかチェックする
                display_result.best(dummy_value_func, i);

                //結果をSCE-UAに返す
                //ファイル名はSCE-UAからもらった番号を使用
                //最後だけ99999を出力しSCE-UA.exeを止める
                using (StreamWriter write_cs2for = new StreamWriter("cmpeeObjVal " + new_para[new_para.Count() - 1] + ".dat"))
                {
                    if (i < cal_count - 1)
                    {
                        write_cs2for.WriteLine(dummy_value_func);
                    }
                    else
                    {
                        write_cs2for.WriteLine("99999");
                    }
                        
                    write_cs2for.Close();
                }

                // 結果の表示(2回目以降)
                list = i + 1 + "          " + true_value_func + "       " + buf_true[i].ToString();
                display_result.textDisplay1(list);

                // 起動させたFortranが消えていないかチェックする <-- 以下は不要。立川 20130114
                //if (cmd_pro_pi.HasExited == true)
                //{
                //    goto Finish;
                //}
            }

            // Finishの表示
            list = "Parameter Identification Finish";
            display_result.textDisplay0(list);


            // 以下のブロックは2013年1月4日に高橋君が追加してくれた。
            // 作業用に用いた SCE-UA.exe の出力ファイルを消去する。 
            string filename_delete;
            for (int i = 0; i < cal_count; i++)
            {
                filename_delete = "cmpeeObjVal " + i.ToString() + ".dat";
                // すでに計算が終わった SCE-UA.exe の出力ファイルを消去する。 
                if (File.Exists(filename_delete))
                {
                    File.Delete(filename_delete);
                }
            }
            // 作業用に用いた出力ファイルを消去する。 
            filename_delete = "cmpeeCount.dat";
            if (File.Exists(filename_delete))
            {
                File.Delete(filename_delete);
            }
            filename_delete = "cmpeeInitParam.dat";
            if (File.Exists(filename_delete))
            {
                File.Delete(filename_delete);
            }
            filename_delete = "cmpeeUpdateParam.dat";
            if (File.Exists(filename_delete))
            {
                File.Delete(filename_delete);
            } 


            // 結果をファイルへ出力
            if (check_out)
            {
                display_result.fileOutput(dataGridView1, columns, rows, ob_func);
            }

            if (cmd_pro_pi.HasExited == false)
            {
                cmd_pro_pi.Kill();
                cmd_pro_pi.Close();
                cmd_pro_pi.Dispose();
            }

        Finish: ;
        }
    }


    class read_calculate
    {
        /// <summary>
        /// 計算を行う(1回分)
        /// </summary>
        /// <param name="filename_pro">プロジェクトファイル名</param>
        /// <param name="hymco_dir">hymco.exeのディレクトリ</param>
        public void calculate(
            string filename_pro, // CommonMP のプロジェクトファイル名
            string hymco_dir)    // hymco.exeのディレクトリ
        {
            Directory.SetCurrentDirectory(hymco_dir); // 現在の作業ディレクトリを hymco_dir に設定する
            Process cmd_pro = new Process();
            cmd_pro.StartInfo.FileName = "cmd.exe"; // コマンドプロンプトの文字列設定
            cmd_pro.StartInfo.Arguments = "/K " + "hymco.exe -c " + filename_pro; // CommonMP の実行の文字列設定

            cmd_pro.StartInfo.UseShellExecute = false;
            cmd_pro.StartInfo.CreateNoWindow = true;
            cmd_pro.StartInfo.RedirectStandardOutput = true;

            cmd_pro.Start(); // コマンドラインでの hymco.exe の実行

            // コマンドプロンプトが作動している間はここで止めておく
            string resulting = cmd_pro.StandardOutput.ReadToEnd();
            //while (resulting != null)
            //{
            //    resulting = cmd_pro.StandardOutput.ReadLine();
            //}
            cmd_pro.WaitForExit();

            // 計算終了後コマンドプロンプトを消す
            if (cmd_pro.HasExited == false) // コマンドプロムプトが動いている場合
            {
                cmd_pro.Kill();
                cmd_pro.Close();
                cmd_pro.Dispose();
            }
        }
    }

    class objective_function
    {
        string time;
        double value_cal;
        double value_obs;
        string line_data;
        ArrayList buf_cal = new ArrayList();
        ArrayList buf_obs = new ArrayList();
        ArrayList buf_cal_time = new ArrayList();
        ArrayList buf_obs_time = new ArrayList();     
        
        public objective_function(string filename_cal, string filename_obs)
        {
            //計算結果読み取り
            if (File.Exists(filename_cal))
            {
                using (StreamReader read_cal = new StreamReader(filename_cal))
                {
                    for (int i = 0; i < 3; i++) // 3行読み飛ばす
                    {
                        line_data = read_cal.ReadLine();
                    }
                    while ((line_data = read_cal.ReadLine()) != null) // 1行読む
                    {
                        if (line_data.Length > 0)
                        {
                            time = line_data.Split(',')[0].ToString(); // カンマ「,」で区切った1つ目のデータ
                            buf_cal_time.Add(time);
                            value_cal = double.Parse(line_data.Split(',')[1].ToString()); // カンマ「,」で区切った2つ目のデータ
                            buf_cal.Add(value_cal);
                        }
                    }
                    read_cal.Close();
                }
            }
            if (File.Exists(filename_obs))
            {
                //観測値読み取り
                using (StreamReader read_obs = new StreamReader(filename_obs))
                {
                    for (int i = 0; i < 3; i++) // 3行読み飛ばす
                    {
                        line_data = read_obs.ReadLine();
                    }
                    while ((line_data = read_obs.ReadLine()) != null) // 1行読む
                    {
                        if (line_data.Length > 0)
                        {
                            time = line_data.Split(',')[0].ToString(); // カンマ「,」で区切った1つ目のデータ
                            buf_obs_time.Add(time);
                            value_obs = double.Parse(line_data.Split(',')[1].ToString()); // カンマ「,」で区切った2つ目のデータ
                            buf_obs.Add(value_obs);
                        }
                    }
                    read_obs.Close();
                }
            }
        }


        /// <summary>
        /// 目的関数の真値と計算用の値を返す(真値の最適値を最大値と一致させるため)
        /// </summary>
        /// <param name="cal_name">目的関数の名前</param>
        /// <param name="time_type">時間のタイプ</param>
        /// <returns></returns>
        public ArrayList ob_cal(string cal_name, string time_type)
        {
            ArrayList value_of_object = new ArrayList();
            // Nash指標
            if (cal_name == "Nash指標")
            {
                double nash = 1.0;
                double nash_numerator = 0.0;        //Nash指標算出式第2項の分子
                double nash_denominator = 0.0;      //Nash指標算出式第2項の分母
                double obs_ave = 0.0;               //観測値の平均

                // 観測値の平均の算出
                for (int i = 0; i < buf_obs.Count; i++)
                {
                    value_obs = double.Parse(buf_obs[i].ToString());
                    obs_ave += value_obs;
                }

                obs_ave = obs_ave / buf_obs.Count;

                // Nash指標算出式第2項の分子・分母の算出
                for (int i = 0; i < buf_obs.Count; i++)
                {
                    // i番目の値をdouble型にする
                    value_obs = double.Parse(buf_obs[i].ToString());
                    value_cal = double.Parse(buf_cal[i].ToString());

                    // 分子・分母の算出
                    nash_numerator += Math.Pow((value_obs - value_cal), 2);
                    nash_denominator += Math.Pow((value_obs - obs_ave), 2);
                }

                nash = 1 - (nash_numerator / nash_denominator);
                
                // 値の格納　
                value_of_object.Add(nash);
                value_of_object.Add(nash);

                return value_of_object;
            }
            // 流量誤差比率
            else if (cal_name == "流量誤差比率")
            {
                double eq = 0.0;
                double max_obs = 0.1;
                value_obs=0.0;
                value_cal=0.0;

                for (int i = 0; i < buf_obs.Count; i++)
                {
                    if (max_obs < double.Parse(buf_obs[i].ToString()))
                    {
                        max_obs = double.Parse(buf_obs[i].ToString());
                    }
                }

                for (int i = 0; i < buf_obs.Count; i++)
                {
                    value_obs = double.Parse(buf_obs[i].ToString());
                    value_cal = double.Parse(buf_cal[i].ToString());
                    eq += Math.Pow((value_obs - value_cal) / max_obs, 2);
                }

                eq = eq / buf_obs.Count;

                // 値の格納　
                value_of_object.Add(eq);
                value_of_object.Add(1/(eq+1));

                return value_of_object;
            }
            // ピーク流量誤差            
            else if (cal_name == "ピーク流量誤差")
            {
                double peak = 1.0;
                double max_cal = 0.1;
                double max_obs = 0.1;

                for (int i = 0; i < buf_cal.Count; i++)
                {
                    if (max_cal < double.Parse(buf_cal[i].ToString()))
                    {
                        max_cal = double.Parse(buf_cal[i].ToString());
                    }
                }
                for (int i = 0; i < buf_obs.Count; i++)
                {
                    if (max_obs < double.Parse(buf_obs[i].ToString()))
                    {
                        max_obs = double.Parse(buf_obs[i].ToString());
                    }
                }

                peak = max_cal / max_obs;
                // dummy_peakは1を最適値となるようにpeakを計算し直している
                double dummy_peak = 1/(Math.Abs(peak-1)+1);

                // 値の格納　
                value_of_object.Add(peak);
                value_of_object.Add(dummy_peak);

                return value_of_object;
            }
            // ピーク生起位相差
            else if (cal_name == "ピーク生起位相差")
            {
                double max=0.0;
                string time_cal="0.0";
                string time_obs="0.0";

                for (int i = 0; i < buf_cal.Count; i++)
                {
                    if (max < double.Parse(buf_cal[i].ToString()))
                    {
                        time_cal = buf_cal_time[i].ToString();
                    }
                }

                max = 0.0;
                for (int i = 0; i < buf_obs.Count; i++)
                {
                    if (max < double.Parse(buf_obs[i].ToString()))
                    {
                        time_obs = buf_obs_time[i].ToString();
                    }
                }

                // time_typeによって計算を変える
                double time_dif =0.0;
                try
                {
                    if (time_type == "s")
                    {
                        time_dif = double.Parse(time_cal) - double.Parse(time_obs);
                    }
                    else
                    {
                        DateTime dtime_cal = DateTime.Parse(time_cal);
                        DateTime dtime_obs = DateTime.Parse(time_obs);
                        time_dif = (dtime_cal - dtime_obs).Days * 3600 * 24 + (dtime_cal - dtime_obs).Hours * 3600 + (dtime_cal - dtime_obs).Minutes * 60
                            + (dtime_cal - dtime_obs).Seconds;
                    }
                }
                catch
                {
                    MessageBox.Show("TimeType is error","Error");
                }

                // 値の格納(分母が0となることを防ぐため1足す)
                value_of_object.Add(time_dif);
                value_of_object.Add(1 / (Math.Abs(time_dif)+1));

                return value_of_object;
            }
            // 流出ボリューム誤差
            else if (cal_name == "流出ボリューム誤差")
            {
                double ev = 0.0;
                value_obs = 0.0;
                value_cal = 0.0;
                for (int i = 0; i < buf_cal.Count; i++)
                {
                    value_cal += double.Parse(buf_cal[i].ToString());
                    value_obs += double.Parse(buf_obs[i].ToString()); 
                }
                ev = (value_cal - value_obs) / value_obs;
                
                //値の格納　
                value_of_object.Add(ev);
                value_of_object.Add(1 / (ev+1));

                return value_of_object;
            }
            else
            {
                return value_of_object;
            }
        }
    }
}
